#include <torch/extension.h>
#include <ATen/ATen.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <vector>

#define CHECK_CUDA(x) TORCH_CHECK(x.is_cuda(), #x " must be a CUDA tensor")
#define CHECK_CONTIGUOUS(x) TORCH_CHECK(x.is_contiguous(), #x " must be contiguous")
#define CHECK_INPUT(x) CHECK_CUDA(x); CHECK_CONTIGUOUS(x)

constexpr int TILE_SIZE = 128;

// Device function for computing dot product
__device__ float dot_product(const float* x, const float* cluster, int D) {
    float sum = 0.0f;
    int tid = threadIdx.x;

    for (int i = tid; i < D; i += TILE_SIZE) {
        sum += x[i] * cluster[i];
    }

    return sum;
}

// Device function for applying batch normalization
__device__ float batch_norm(float val, float mean, float var, float weight, float bias) {
    return (val - mean) * weight / sqrtf(var + 1e-5f) + bias;
}

// Device function for applying softmax
__device__ float softmax_reduction(float val, const float* row_cache, int KplusG) {
    float max_val = -INFINITY;
    float sum_exp = 0.0f;

    for (int i = 0; i < KplusG; ++i) {
        max_val = fmaxf(max_val, row_cache[i]);
    }

    __syncthreads();

    val = __expf(val - max_val);

    for (int i = 0; i < KplusG; ++i) {
        sum_exp += __expf(row_cache[i] - max_val);
    }

    __syncthreads();

    return val / sum_exp;
}

__global__ void modular_fused_assignment_kernel(
    const float* __restrict__ x,
    const float* __restrict__ clusters,
    const float* bn_weight,
    const float* bn_bias,
    const float* bn_mean,
    const float* bn_var,
    float* output,
    int64_t BxN,
    int64_t D,
    int64_t KplusG,
    bool is_training) {

    int row = blockIdx.x * blockDim.y + threadIdx.y;
    int col = threadIdx.y;

    __shared__ float smem_cache[TILE_SIZE * TILE_SIZE];

    if (row >= BxN) return;

    extern __shared__ float shared[];
    float* row_cache = shared;

    // Compute dot product
    float dot = dot_product(x + row * D, clusters + col * D, D);
    row_cache[col] = dot;

    __syncthreads();

    // Apply BN, if in inference mode
    if (!is_training) {
        row_cache[col] = batch_norm(row_cache[col], bn_mean[col], bn_var[col], bn_weight[col], bn_bias[col]);
    }

    __syncthreads();

    // Apply softmax
    output[row * KplusG + col] = softmax_reduction(row_cache[col], row_cache, KplusG);
}

void call_modular_fused_assignment_kernel(
    torch::Tensor x,
    torch::Tensor clusters,
    torch::Tensor bn_weight,
    torch::Tensor bn_bias,
    torch::Tensor bn_running_mean,
    torch::Tensor bn_running_var,
    torch::Tensor assignment,
    int64_t BxN,
    int64_t D,
    int64_t KplusG,
    bool is_training) {

    dim3 block(TILE_SIZE, TILE_SIZE);
    dim3 grid((BxN + TILE_SIZE - 1) / TILE_SIZE);
    size_t shared_mem = KplusG * sizeof(float);

    modular_fused_assignment_kernel<<<grid, block, shared_mem>>>(
        x.data_ptr<float>(),
        clusters.data_ptr<float>(),
        bn_weight.data_ptr<float>(),
        bn_bias.data_ptr<float>(),
        bn_running_mean.data_ptr<float>(),
        bn_running_var.data_ptr<float>(),
        assignment.data_ptr<float>(),
        BxN,
        D,
        KplusG,
        is_training);

    cudaDeviceSynchronize();
}

// Forward function
// Refactor code to invoke modular kernel call

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor clusters,
    torch::Tensor clusters2,
    torch::Tensor bn_weight,
    torch::Tensor bn_bias,
    torch::Tensor bn_running_mean,
    torch::Tensor bn_running_var,
    int64_t feature_size,
    int64_t cluster_size,
    bool is_training) {

    CHECK_INPUT(x);
    CHECK_INPUT(clusters);
    CHECK_INPUT(clusters2);
    CHECK_INPUT(bn_weight);
    CHECK_INPUT(bn_bias);
    CHECK_INPUT(bn_running_mean);
    CHECK_INPUT(bn_running_var);

    int64_t B = x.size(0);
    int64_t N = x.size(1);
    int64_t D = feature_size;
    int64_t K = cluster_size;
    int64_t KplusG = clusters.size(1);
    int64_t BxN = B * N;

    x = x.reshape({-1, D});
    auto assignment = torch::empty({BxN, KplusG}, x.options());

    call_modular_fused_assignment_kernel(
        x,
        clusters,
        bn_weight,
        bn_bias,
        bn_running_mean,
        bn_running_var,
        assignment,
        BxN,
        D,
        KplusG,
        is_training);

    assignment = assignment.narrow(1, 0, K).reshape({B, N, K});

    auto a_sum = assignment.sum(1, true);
    clusters2 = clusters2.expand({B, D, K});
    auto a = clusters2 * a_sum;

    assignment = assignment.transpose(1, 2);
    x = x.reshape({B, N, D});
    auto vlad = torch::bmm(assignment, x).transpose(1, 2) - a;

    vlad = torch::nn::functional::normalize(
        vlad, torch::nn::functional::NormalizeFuncOptions().p(2).dim(1));
    vlad = vlad.reshape({B, D * K});
    vlad = torch::nn::functional::normalize(
        vlad, torch::nn::functional::NormalizeFuncOptions().p(2).dim(1));

    return vlad;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "NetVLAD forward");
}
